/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _SCAN_MATCH_HPP
#define _SCAN_MATCH_HPP

// scan-matching RBPF that produces occupancy maps

#include "slam.hpp"

class grid_map_t
{
protected:
  double xmin, ymin, xmax, ymax, res, width, height;
  uint32_t rows, cols, size;
  int8_t *grid;

public:
  grid_map_t();
  grid_map_t(const grid_map_t &m) : grid(0) { *this = m; }
  ~grid_map_t() { if(grid) delete [] grid; grid = 0; }
  grid_map_t & operator=(const grid_map_t &m);
  double get_likelihood_and_update
    (const cartesian_scan_t &points, const pose_t &pose);
  double grid_map_t::get_likelihood(const cartesian_scan_t &points);
  void render() const;
  void write_to_file(const char *prefix) const;

protected:
  inline void rowcol(double x, double y, uint32_t &r, uint32_t &c);
  inline uint32_t cell(uint32_t r, uint32_t c);
  inline void occupied(uint32_t cell);
  inline void unoccupied(uint32_t cell);
  inline double get_likelihood
    (const point_t &p, uint32_t r, uint32_t c,
     point_t &gp, double &d);
  inline void match_single_point
    (const point_t &p, double &likelihood, uint32_t &r, uint32_t &c);
  void grid_map_t::bresenham(int x0, int y0, int x1, int y1);
  void update(const point_t &p, const pose_t &pose);
};

struct grid_particle_t : public particle_t
{
  grid_map_t grid_map;
};

class slam_scan_match
{
public:
  slam_scan_match(const pose_t &start_pose = pose_t());
  void add_odometry(const pose_t &delta, const pose_t &ref_frame);
  void add_scan(const multiscan_t &points);
  void render();
  void write_to_file(const char *prefix) const;
protected:
  std::vector<grid_particle_t> particles;
  uint32_t best; // best particle

  // most recent scan and extracted features
  pose_t frame;
  cartesian_scan_t all_points;
};

// with improved proposal distribution (use the current measurement
// when drawing particles --- fastslam 2, basically)
class slam_scan_match_2 : public slam_scan_match
{
public:
  slam_scan_match_2(const pose_t &start_pose = pose_t())
    : slam_scan_match(start_pose) {}
  void add_scan(const multiscan_t &points);
  void add_odometry(const pose_t &delta, const pose_t &ref_frame);
protected:
  pose_t last_delta, last_ref_frame;
};

#endif // _SCAN_MATCH_HPP
